import os
import numpy as np
import galsim
from .base_object import BaseObject

__all__ = ['StarObject']

class StarObject(BaseObject):
    _type_name = 'star'
    def _get_sed(self, mjd=None, redshift=0):
        '''
        We'll need mjd when/if variable stars are supported. For now
        it's ignored.
        '''
        mag_norm = self.get_native_attribute('magnorm')
        rel_path = self.get_native_attribute('sed_filepath')

        fpath = os.path.join(os.getenv('SIMS_SED_LIBRARY_DIR'),
                             self.get_native_attribute('sed_filepath'))

        return self._get_sed_from_file(fpath), mag_norm

    def get_gsobject_components(self, gsparams=None, rng=None):
        if gsparams is not None:
            gsparams = galsim.GSParams(**gsparams)
        return {'this_object': galsim.DeltaFunction(gsparams=gsparams)}

    def get_observer_sed_component(self, component, mjd=None):
        sed, magnorm = self._get_sed(mjd=mjd)

        # The SED is in units of photons/nm/cm^2/s
        # -0.9210340371976184 = -np.log(10)/2.5. Use to convert mag to flux
        flux_500 = np.exp(-0.9210340371976184 * magnorm)
        sed = sed.withMagnitude(0, self._bp500)
        sed = sed*flux_500

        if sed is not None:
            sed = self._apply_component_extinction(sed)
        return sed



__all__ = ['CatalogContext']
class CatalogContext:
    def __init__(self, the_sky_cat):
        global sky_cat
        self._sky_cat = the_sky_cat
        sky_cat = the_sky_cat
        source_type_dict = {}

        # Initialize with known source types and corresponding object classes
        # Only include collection class if it's not "standard"
        # (that is, not ObjectCollection)
        # source_type_dict['star'] = {'object_class' : StarObject}
        # source_type_dict['sncosmo'] = {'object_class' : SncosmoObject}
        # source_type_dict['galaxy'] = {'object_class' : GalaxyObject}
        self._source_type_dict = source_type_dict

    def register_source_type(self, name, object_class, collection_class=None):
        self._source_type_dict[name] = {'object_class' : object_class,
                                        'collection_class' : collection_class}

    def lookup_source_type(self, name):
        if name in self._source_type_dict:
            return self._source_type_dict[name]['object_class']
        else:
            return None

    def lookup_collection_type(self, name):
        if name in self._source_type_dict:
            if 'collection_class' in self._source_type_dict[name]:
                return self._source_type_dict[name]['collection_class']
            else:
                return None
        else:
            return None



from .common_utils import *
from .config_utils import *
from .translate_utils import *
from .exceptions import *
from .parquet_schema_utils import *
from .sed_tools import *
from .sn_tools import *
from .shapes import *



import numpy as np
from astropy import units as u
import sncosmo
import galsim

__all__ = ['SncosmoModel']


class SncosmoModel(sncosmo.Model):
    def __init__(self, source='salt2-extended', params=None):
        '''
        params - dict of params suitable for the model

        See also https://sncosmo.readthedocs.io/en/stable/index.html

        '''
        # The following explicitly turns off host and Milky Way extinction.
        dust = sncosmo.F99Dust()
        super().__init__(source=source,
                         effects=[dust, dust],
                         effect_names=['host', 'mw'],
                         effect_frames=['rest', 'obs'])
        self.set(mwebv=0., hostebv=0.)
        self.redshift = 0
        if params:
            self.set(**params)
            self.redshift = params['z']

    def get_sed(self, mjd, npts=1000):
        """
        Return the SED in the observer frame at the requested time.
        """
        wl = np.linspace(self.minwave(), self.maxwave(), npts)

        # prepend 0 bins
        n_bins = int(self.minwave()) - 1
        pre_wl = [float(i) for i in range(n_bins)]
        pre_val = [0.0 for i in range(n_bins)]
        flambda = self.flux(mjd, wl)
        wl = np.insert(wl, 0, pre_wl)
        flambda = np.insert(flambda, 0, pre_val)
        lut = galsim.LookupTable(wl, flambda, interpolant='linear')
        return galsim.SED(lut, wave_type=u.Angstrom, flux_type='flambda')



import numpy as np
import galsim

from .base_object import BaseObject

__all__ = ['GalaxyObject']

class GalaxyObject(BaseObject):
    _type_name = 'galaxy'
    def _get_sed(self, component=None, resolution=None):
        '''
        Return sed and mag_norm for a galaxy component or for a star
        Parameters
        ----------
        component    one of 'bulge', 'disk', 'knots' for now. Other components
                     may be supported.  Ignored for stars
        resolution   desired resolution of lambda in nanometers. Ignored
                     for stars.

        Returns
        -------
        A pair (sed, mag_norm)
        '''
        if component not in ['disk', 'bulge', 'knots']:
            raise ValueError(f'Cannot fetch SED for component type {component}')

        th_val = self.get_native_attribute(f'sed_val_{component}')
        if  th_val is None:   #  values for this component are not in the file
            raise ValueError(f'{component} not part of this catalog')

        # if values are all zeros or nearly no point in trying to convert
        if max(th_val) < np.finfo('float').resolution:
            return None, 0.0

        z_h = self.get_native_attribute('redshift_hubble')
        z = self.get_native_attribute('redshift')

        sky_cat = self._belongs_to._sky_catalog
        sed = sky_cat.observed_sed_factory.create(th_val, z_h, z,
                                                  resolution=resolution)
        magnorm = sky_cat.observed_sed_factory.magnorm(th_val, z_h)

        return sed, magnorm

    def get_wl_params(self):
        """Return the weak lensing parameters, g1, g2, mu."""
        gamma1 = self.get_native_attribute('shear_1')
        gamma2 = self.get_native_attribute('shear_2')
        kappa =  self.get_native_attribute('convergence')
        # Compute reduced shears and magnification.
        g1 = gamma1/(1. - kappa)    # real part of reduced shear
        g2 = gamma2/(1. - kappa)    # imaginary part of reduced shear
        mu = 1./((1. - kappa)**2 - (gamma1**2 + gamma2**2)) # magnification
        return g1, g2, mu

    def get_total_observer_sed(self, mjd=None):
        """
        Return the SED summed over SEDs for the individual SkyCatalog
        components.
        """
        sed = super().get_total_observer_sed()

        if sed is None:
            return sed

        _, _, mu = self.get_wl_params()
        sed *= mu
        return sed


    def get_gsobject_components(self, gsparams=None, rng=None):

        if gsparams is not None:
            gsparams = galsim.GSParams(**gsparams)

        obj_dict = {}
        for component in self.subcomponents:
            # knots use the same major/minor axes as the disk component.
            my_component = 'disk' if component != 'bulge' else 'bulge'
            a = self.get_native_attribute(
                f'size_{my_component}_true')
            b = self.get_native_attribute(
                f'size_minor_{my_component}_true')
            assert a >= b
            hlr = (a*b)**0.5   # approximation for half-light radius

            e1 = self.get_native_attribute(
                f'ellipticity_1_{my_component}_true')
            e2 = self.get_native_attribute(
                f'ellipticity_2_{my_component}_true')

            if component == 'knots':
                npoints = self.get_native_attribute('n_knots')
                assert npoints > 0
                obj = galsim.RandomKnots(npoints=npoints,
                                         half_light_radius=hlr, rng=rng,
                                         gsparams=gsparams)
            else:
                n = self.get_native_attribute(f'sersic_{component}')
                # Quantize the n values at 0.05 so that galsim can
                # possibly amortize sersic calculations from the previous
                # galaxy.
                n = round(n*20.)/20.
                obj = galsim.Sersic(n=n, half_light_radius=hlr,
                                    gsparams=gsparams)

            # NOTE: The minus signs in the next executable line are
            # needed specifically for generating DC2-like from galaxy
            # catalogs in the cosmoDC2 format. They are included here in
            # order to reproduce the effect of adding 90 degrees to
            # position angle in the old code. Newer input galaxy catalogs
            # most likely will not need this adjustment
            shear = galsim.Shear(g1=-e1, g2=-e2)
            obj = obj._shear(shear)
            g1, g2, mu = self.get_wl_params()
            obj_dict[component] = obj._lens(g1, g2, mu)
        return obj_dict

    def get_observer_sed_component(self, component, mjd=None):
        # For now make SED with resolution of 1 nm
        sed, _ = self._get_sed(component=component, resolution=1.0)
        if sed is not None:
            sed = self._apply_component_extinction(sed)

        return sed

    def get_instcat_entry(self, band = 'r', component=None):
        '''
        Return the string corresponding to instance catalog line
        Parameters:
            band       One of ['u', 'g', 'r', 'i', 'z', 'y']
            component  Required iff the object has subcomponents (i.e.,
                       object type is 'galaxy')
        Returns: A string formatted like a line in an instance catalog
        '''
        if component not in self.subcomponents:
            return ''
        return form_object_string(self, band, component)



import os
import sys
import warnings
import itertools
from pathlib import PurePath
import numpy as np
import erfa
import astropy.modeling
from astropy import units as u
import galsim
import lsst.daf.butler as daf_butler
import lsst.geom
from lsst.meas.algorithms import ReferenceObjectLoader
from skycatalogs.utils.shapes import Disk, PolygonalRegion
from skycatalogs.objects.base_object import BaseObject, ObjectCollection


__all__ = ['GaiaObject', 'GaiaCollection']


_FILE_PATH = str(PurePath(__file__))
_SKYCATALOGS_DIR = _FILE_PATH[:_FILE_PATH.rindex('/skycatalogs')]


# Read in function for stellar temperature estimation given bp and rp
# fluxes
_TEMP_EST_DATA_FILE = os.path.join(_SKYCATALOGS_DIR, 'skycatalogs',
                                   'data', 'gaia_dr2',
                                   'gaia_dr2_temp_from_bp-rp_ratio.txt')
_TEMP_FUNC = galsim.LookupTable(*np.genfromtxt(_TEMP_EST_DATA_FILE))

# Read in the Gaia DR2 passband for the bp flux to use for setting the
# SED normalizations.
_GAIA_PB_FILE = os.path.join(_SKYCATALOGS_DIR, 'skycatalogs',
                             'data', 'gaia_dr2',
                             'GaiaDR2_RevisedPassbands.dat')
_GAIA_PBS = np.genfromtxt(_GAIA_PB_FILE, names=['wl', 'g', 'g_err', 'bp',
                                                'bp_err', 'rp', 'rp_err'])
_index = np.where(_GAIA_PBS['bp'] < 99)  # Omit sentinel values.
# Create the Bandpass object for the bp passband, using the 'AB' zero point.
_GAIA_BP = galsim.Bandpass(
    galsim.LookupTable(_GAIA_PBS['wl'][_index], _GAIA_PBS['bp'][_index]),
    wave_type='nm').withZeropoint('AB').thin()


class GaiaObject(BaseObject):
    """
    Object class for Gaia DR2 data read from the LSST reference catalog files.
    """
    _stellar_temperature = _TEMP_FUNC
    _gaia_bp_bandpass = _GAIA_BP
    _wavelengths = np.arange(250, 1250, 5, dtype=float)
    def __init__(self, obj_pars, parent_collection, index):
        """
        Parameters
        ----------
        obj_pars: dict-like
            Dictionary of object parameters as read directly from the
            reference catalog.
        parent_collection: ObjectCollection
            Parent collection of this object.
        index: int
            Index of this object in the parent collection
        use_lut: bool [True]
            Flag to use a galsim LookupTable for representing the SED.
        """
        ra = np.degrees(obj_pars['coord_ra'])
        dec = np.degrees(obj_pars['coord_dec'])
        # Form the object id from the GAIA catalog id with the string
        # 'gaia_dr2_' prepended.
        obj_id = f"gaia_dr2_{obj_pars['id']}"
        super().__init__(ra, dec, obj_id, 'gaia_star',
                         belongs_to=parent_collection, belongs_index=index)
        self.use_lut = self._belongs_to._use_lut
        bp_flux = obj_pars['phot_bp_mean_flux']
        rp_flux = obj_pars['phot_rp_mean_flux']
        if rp_flux == 0.0 or bp_flux == 0.0:
            self.stellar_temp = None
        else:
            try:
                self.stellar_temp = self._stellar_temperature(bp_flux/rp_flux)
                # Convert from flux units of nJy to AB mag for the bp passband,
                # which we will use to normalize the SED.
                self.bp_mag = -2.5*np.log10(bp_flux*1e-9) + 8.90
            except galsim.errors.GalSimRangeError as ex:
                self.stellar_temp = None
            except RuntimeError as rex:
                self.stellar_temp = None

    def blambda(self, wl):
        clight = astropy.constants.c.value*1e2  # cm/s
        nu = clight/(wl*1e-7)  # Hz
        Bnu = astropy.modeling.physical_models.BlackBody(
            temperature=self.stellar_temp*u.K)
        return Bnu(nu*u.Hz).value*nu**2/clight/1e7  # erg/nm/cm^2/s

    def get_observer_sed_component(self, component, mjd=None):
        if component != 'this_object':
            raise RuntimeError("Unknown SED component: %s", component)
        if self.stellar_temp is None:
            return None
        if self.use_lut:
            flambda = self.blambda(self._wavelengths)
            lut = galsim.LookupTable(self._wavelengths, flambda)
            sed = galsim.SED(lut, wave_type='nm', flux_type='flambda')
        else:
            sed = galsim.SED(self.blambda, wave_type='nm', flux_type='flambda')
        return sed.withMagnitude(self.bp_mag, self._gaia_bp_bandpass)

    def get_gsobject_components(self, gsparams=None, rng=None):
        if gsparams is not None:
            gsparams = galsim.GSParams(**gsparams)
        return {'this_object': galsim.DeltaFunction(gsparams=gsparams)}

    def set_use_lut(self, use_lut):
        self.use_lut = use_lut

class GaiaCollection(ObjectCollection):
    # Class methods
    _gaia_config = None
    def set_config(config):
        GaiaCollection._gaia_config = config

    def get_config():
        return GaiaCollection._gaia_config

    def load_collection(region, skycatalog, mjd=None):
        if isinstance(region, Disk):
            ra = lsst.geom.Angle(region.ra, lsst.geom.degrees)
            dec = lsst.geom.Angle(region.dec, lsst.geom.degrees)
            center = lsst.geom.SpherePoint(ra, dec)
            radius = lsst.geom.Angle(region.radius_as, lsst.geom.arcseconds)
            refcat_region = lsst.sphgeom.Circle(center.getVector(), radius)
        elif isinstance(region, PolygonalRegion):
            refcat_region = region._convex_polygon
        else:
            raise TypeError(f'GaiaCollection.load_collection: {region} not supported')

        source_type = 'gaia_star'

        if GaiaCollection.get_config() is None:
            gaia_section = skycatalog.raw_config['object_types']['gaia_star']
            GaiaCollection.set_config(gaia_section)

        butler_params = GaiaCollection.get_config()['butler_parameters']
        butler = daf_butler.Butler(butler_params['repo'],
                                   collections=butler_params['collections'])
        refs = set(butler.registry.queryDatasets(butler_params['dstype']))
        refCats = [daf_butler.DeferredDatasetHandle(butler, _, {})
                   for _ in refs]
        dataIds = [butler.registry.expandDataId(_.dataId) for _ in refs]
        config = ReferenceObjectLoader.ConfigClass()
        config.filterMap = {f'{_}': f'phot_{_}_mean' for _ in ('g', 'bp', 'rp')}
        ref_obj_loader = ReferenceObjectLoader(dataIds=dataIds,
                                               refCats=refCats,
                                               config=config)

        sed_method = GaiaCollection.get_config().get('sed_method', 'use_lut')
        use_lut = (sed_method.strip().lower() == 'use_lut')
        band = 'bp'
        cat = ref_obj_loader.loadRegion(refcat_region, band).refCat
        df =  cat.asAstropy().to_pandas()

        if mjd is None:
            raise RuntimeError("MJD needs to be provided for Gaia "
                               "proper motion calculations.")
        # The erfa.pmsafe code
        # (https://pyerfa.readthedocs.io/en/latest/api/erfa.pmsafe.html)
        # expects ra, dec in units of radians and pm_ra, pm_dec
        # (proper motion) in units of radians/year.  These are the
        # units used in the refcats files.  However, erfa.pmsafe
        # expects parallax in arcseconds so convert from the refcats
        # units of radians:
        arcsec_per_radian = (1.0*u.radian).to(u.arcsec).value
        df['parallax'] *= arcsec_per_radian
        # radial velocity is missing from the Gaia DR2 refcats, so pass
        # an array of zeros.
        rv1 = np.zeros(len(df))
        epNa = 2400000.5  # "part A" of starting and ending epochs for MJDs.
        ep2b = mjd
        pm_outputs = erfa.pmsafe(df['coord_ra'], df['coord_dec'],
                                 df['pm_ra'], df['pm_dec'], df['parallax'],
                                 rv1, epNa, df['epoch'], epNa, ep2b)
        # Update ra, dec values.
        df['coord_ra'] = pm_outputs[0]
        df['coord_dec'] = pm_outputs[1]

        return GaiaCollection(df, skycatalog, source_type, use_lut, mjd)

    def __init__(self, df, sky_catalog, source_type, use_lut, mjd):
        self.df = df
        self._sky_catalog = sky_catalog
        self._partition_id = None
        self._id = np.array([f"gaia_dr2_{df.iloc[key]['id']}" for key in range(len(df))])
        self._mask = None
        self._object_type_unique = source_type
        self._use_lut = use_lut
        self._rdrs = []
        self._object_class = GaiaObject
        self._use_lut = use_lut
        self._mjd = mjd

    @property
    def native_columns(self):
        return set()

    @property
    def use_lut(self):
        return self._use_lut

    @property
    def mjd(self):
        return self._mjd

    def __getitem__(self, key):
        if isinstance(key, int) or isinstance(key, np.int64):
            return GaiaObject(self.df.iloc[key], self, key)

        elif type(key) == slice:
            ixdata = [i for i in range(min(key.stop,len(self._id)))]
            ixes = itertools.islice(ixdata, key.start, key.stop, key.step)
            return [self._object_class(self.df.iloc[i], self, i) for i in ixes]

        elif type(key) == tuple and isinstance(key[0], Iterable):
            #  check it's a list of int-like?
            return [self._object_class(self.df.iloc[i], self, i) for i in key[0]]

    def __len__(self):
        return len(self.df)



import galsim
from skycatalogs.utils.sn_tools import SncosmoModel
from .base_object import BaseObject


__all__ = ['SncosmoObject']

class SncosmoObject(BaseObject):
    _type_name = 'sncosmo'
    def _get_sed(self, mjd=None):
        params = self.get_native_attribute('salt2_params')
        sn = SncosmoModel(params=params)
        if mjd < sn.mintime() or mjd > sn.maxtime():
            return None, 0.0
        # Already normalized so magnorm is zero
        return sn.get_sed(mjd), 0.0

    def get_gsobject_components(self, gsparams=None, rng=None):
        if gsparams is not None:
            gsparams = galsim.GSParams(**gsparams)
        return {'this_object': galsim.DeltaFunction(gsparams=gsparams)}

    def get_observer_sed_component(self, component, mjd=None):
        sed, _ = self._get_sed(mjd=mjd)
        if sed is not None:
            sed = self._apply_component_extinction(sed)
        return sed

    def get_LSST_flux(self, band, sed=None, mjd=None):
        if not band in LSST_BANDS:
            return None

        return self.get_flux(lsst_bandpasses[band], sed=sed, mjd=mjd)



from collections import namedtuple
import numpy as np
from astropy import units as u


__all__ = ['Box', 'Disk', 'PolygonalRegion']

Box = namedtuple('Box', ['ra_min', 'ra_max', 'dec_min', 'dec_max'])

# radius is measured in arcseconds
Disk = namedtuple('Disk', ['ra', 'dec', 'radius_as'])

from lsst.sphgeom import ConvexPolygon, UnitVector3d, LonLat
class PolygonalRegion:

    def __init__(self, vertices_radec=None, convex_polygon=None):
        '''
        Supply either an object of type lsst.sphgeom.ConvexPolygon
        or a list of vertices, each a tuple (ra, dec) in degrees,
        which describe a convex polygon
        '''
        if convex_polygon:
            if isinstance(convex_polygon, ConvexPolygon):
                self._convex_polygon = convex_polygon
                return
        if vertices_radec:
            if not isinstance(vertices_radec, list):
                raise TypeError(f'PolygonalRegion: Argument {vertices_radec} is not a list')
            vertices = [UnitVector3d(LonLat.fromDegrees(v_rd[0], v_rd[1])) for v_rd in vertices_radec]
            self._convex_polygon = ConvexPolygon(vertices)
            return
        raise ValueError('PolygonalRegion: Either vertices_radec or convex_polygon must have an acceptable value')

    def get_vertices(self):
        '''
        Return vertices as list of 3d vectors
        '''
        return self._convex_polygon.getVertices()

    def get_vertices_radec(self):
        v3d = self.get_vertices()
        vertices_radec = []
        for v in v3d:
            vertices_radec.append((LonLat.longitudeOf(v).asDegrees(),
                                   LonLat.latitudeOf(v).asDegrees()))
        return vertices_radec

    def get_containment_mask(self, ra, dec, included=True):
        '''
        Parameters
        ----------
        ra, dec      parallel float arrays, units are degrees. Together
                     they define the list of points to be checked for
                     containment
        included     boolean   If true, mask bit will be set to True for
                               contained points, else False.    Reverse
                               the settings if included is False
        '''
        # convert to radians
        ra = [(r * u.degree).to_value(u.radian) for r in ra]
        dec = [(d * u.degree).to_value(u.radian) for d in dec]

        mask = self._convex_polygon.contains(ra, dec)
        if included:
            return mask
        else:
            return np.logical_not(mask)



import os
import sys
import re
import yaml
import logging
import healpy
import numpy as np
import numpy.ma as ma
import pyarrow.parquet as pq
from astropy import units as u
from skycatalogs.objects.base_object import load_lsst_bandpasses
from skycatalogs.utils.catalog_utils import CatalogContext
from skycatalogs.objects.base_object import ObjectList, ObjectCollection
from skycatalogs.objects.gaia_object import GaiaObject, GaiaCollection
from skycatalogs.readers import ParquetReader
from skycatalogs.utils.sed_tools import TophatSedFactory
from skycatalogs.utils.sed_tools import MilkyWayExtinction
from skycatalogs.utils.config_utils import Config
from skycatalogs.utils.shapes import Box, Disk, PolygonalRegion
from skycatalogs.objects.sncosmo_object import SncosmoObject
from skycatalogs.objects.star_object import StarObject
from skycatalogs.objects.galaxy_object import GalaxyObject

__all__ = ['SkyCatalog', 'open_catalog']

# This function should maybe be moved to utils
def _get_intersecting_hps(hp_ordering, nside, region):
    '''
    Given healpixel structure defined by hp_ordering and nside, find
    all healpixels which intersect region, where region may be a box,
    a disk or a polygonal region.
    Return as some kind of iterable
    Note it's possible extra hps which don't actually intersect the region
    will be returned
    '''
    # First convert region description to an array (4,3) with (x,y,z) coords
    # for each vertex
    if isinstance(region, Box):
        vec = healpy.pixelfunc.ang2vec([region.ra_min, region.ra_max,
                                        region.ra_max, region.ra_min],
                                       [region.dec_min, region.dec_min,
                                        region.dec_max, region.dec_max],
                                       lonlat=True)

        return healpy.query_polygon(nside, vec, inclusive=True, nest=False)
    if isinstance(region, Disk):
        # Convert inputs to the types query_disk expects
        center = healpy.pixelfunc.ang2vec(region.ra, region.dec,
                                          lonlat=True)
        radius_rad = (region.radius_as * u.arcsec).to_value('radian')

        pixels = healpy.query_disc(nside, center, radius_rad, inclusive=True,
                                   nest=False)
        return pixels

    if isinstance(region, PolygonalRegion):
        return healpy.query_polygon(nside, region.get_vertices(),
                                    inclusive=True, nest=False)

def _compress_via_mask(tbl, id_column, region, galaxy=True):
    '''
    Parameters
    ----------
    tbl          data table including columns named "ra", "dec", and id_column
                 (and also "object_type" colum if galaxy is False)
    id_column    string
    region       mask should restrict to this region (or not at all if None)
    galaxy       flag so we know whether or not to return "object_type"

    Returns
    -------
    4 values for galaxies: ra, dec, id, mask
    5 values for pointsources: ra, dec, id, object_type, mask
    If objects are in the region, ra, dec, id correspond to those objects.
    mask will mask off unused objects
    If there are no objects in the region, all return values are None

    '''
    if region is not None:
        if isinstance(region, PolygonalRegion):        # special case
            # Find bounding box
            vertices = region.get_vertices_radec()  # list of (ra, dec)
            ra_max = max([v[0] for v in vertices])
            ra_min = min([v[0] for v in vertices])
            dec_max = max([v[1] for v in vertices])
            dec_min = min([v[1] for v in vertices])
            bnd_box = Box(ra_min, ra_max, dec_min, dec_max)
            # Compute mask for that box
            mask = _compute_mask(bnd_box, tbl['ra'], tbl['dec'])
            if all(mask): # even bounding box doesn't intersect table rows
                if galaxy:
                    return None, None, None, None
                else:
                    return None, None, None, None, None

            # Get compressed ra, dec
            ra_compress = ma.array(tbl['ra'], mask=mask).compressed()
            dec_compress = ma.array(tbl['dec'], mask=mask).compressed()
            poly_mask = _compute_mask(region, ra_compress, dec_compress)

            # Get indices of unmasked
            ixes = np.where(~mask)

            # Update bounding box mask with polygonal mask
            mask[ixes] |= poly_mask
        else:
            mask = _compute_mask(region, tbl['ra'], tbl['dec'])

        if all(mask):
            if galaxy:
                return None, None, None, None
            else:
                return None, None, None, None, None
        else:
            ra_compress = ma.array(tbl['ra'], mask=mask).compressed()
            dec_compress = ma.array(tbl['dec'], mask=mask).compressed()
            id_compress = ma.array(tbl[id_column], mask=mask).compressed()
            if galaxy:
                return ra_compress, dec_compress, id_compress, mask
            else:
                object_type_compress = ma.array(tbl['object_type'],
                                                mask=mask).compressed()
                return ra_compress, dec_compress, id_compress, object_type_compress, mask
    else:
        if galaxy:
            return tbl['ra'], tbl['dec'], tbl[id_column],None
        else:
            return tbl['ra'], tbl['dec'], tbl[id_column],tbl['object_type'],None

def _compute_mask(region, ra, dec):
    '''
    Compute mask according to region for provided data
    Parameters
    ----------
    region         Supported shape (box, disk)  or None
    ra,dec         Coordinates for data to be masked, in degrees
    Returns
    -------
    mask of elements in ra, dec arrays to be omitted

    '''
    mask = None
    if isinstance(region, Box):
        mask = np.logical_or((ra < region.ra_min),
                             (ra > region.ra_max))
        mask = np.logical_or(mask, (dec < region.dec_min))
        mask = np.logical_or(mask, (dec > region.dec_max))
    if isinstance(region, Disk):
        # Change positions to 3d vectors to measure distance
        p_vec = healpy.pixelfunc.ang2vec(ra, dec, lonlat=True)

        c_vec = healpy.pixelfunc.ang2vec(region.ra,
                                         region.dec,
                                         lonlat=True)
        radius_rad = (region.radius_as * u.arcsec).to_value('radian')


        # Rather than comparing arcs, it is equivalent to compare chords
        # (or square of chord length)
        diff = p_vec - c_vec
        obj_chord_sq = np.sum(np.square(p_vec - c_vec),axis=1)

        # This is to be compared to square of chord for angle a corresponding
        # to disk radius.  That's 4(sin(a/2)^2)
        rad_chord_sq = 4 * np.square(np.sin(0.5 * radius_rad) )
        mask = obj_chord_sq > rad_chord_sq
    if isinstance(region, PolygonalRegion):
        mask = region.get_containment_mask(ra, dec, included=False)
    return mask

class SkyCatalog(object):
    '''
    A base class with derived classes for galaxies, static (w.r.t. coordinates)
    point sources, SSOs

    '''
    def __init__(self, config, mp=False, skycatalog_root=None, verbose=False,
                 loglevel='INFO'):
        '''
        Parameters
        ----------
        config:  dict.  Typically the result of loading a yaml file
        mp:      boolean  Default False. Set True to enable multiprocessing.
        skycatalog_root: If not None, overrides value in config or
                         in environment variable SKYCATALOG_ROOT
        '''
        self._config = Config(config)
        self._global_partition = None
        if 'area_partition' in self._config.keys():       # old-style config
            self._global_partition = self._config['area_partition']

        self._logger = logging.getLogger('skyCatalogs.client')

        if not self._logger.hasHandlers():
            self._logger.setLevel(loglevel)
            ch = logging.StreamHandler()
            ch.setLevel(loglevel)
            formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
            ch.setFormatter(formatter)
            self._logger.addHandler(ch)

        self._mp = mp
        if 'schema_version' not in config and 'schema_version' not in config['provenance']['versioning']:
            self._cat_dir = config['root_directory']
        else:
            sky_root = config['skycatalog_root']        # default
            if skycatalog_root:
                sky_root = skycatalog_root
            else:
                sky_root_env = os.getenv('SKYCATALOG_ROOT', None)
                if sky_root_env:
                    sky_root = sky_root_env

            self._cat_dir = os.path.join(sky_root, config['catalog_dir'])

        self._logger.info(f'Catalog data will be read from {self._cat_dir}')
        # There may be more to do at this point but not too much.
        # In particular, don't read in anything from data files

        self.verbose = verbose
        self._validate_config()


        # Outer dict: hpid for key. Value is another dict
        #    with keys 'files', 'object_types', each with value another dict
        #       for 'files', map filepath to handle (initially None)
        #       for 'object_types', map object type to filepath
        self._hp_info = dict()
        hps = self._find_all_hps()

        # NOTE: the use of TophatSedFactory is appropriate *only* for an
        # input galaxy catalog with format like cosmoDC2, which includes
        # definitions of tophat SEDs. A different implementation will
        # be needed for newer galaxy catalogs
        th_parameters = self._config.get_tophat_parameters();
        self._observed_sed_factory =\
            TophatSedFactory(th_parameters, config['Cosmology'])

        self._extinguisher = MilkyWayExtinction()

        # Make our properties accessible to BaseObject, etc.
        self.cat_cxt = CatalogContext(self)

        # register object types which are in the config
        if 'gaia_star' in config['object_types']:
            self.cat_cxt.register_source_type('gaia_star',
                                              object_class=GaiaObject,
                                              collection_class=GaiaCollection)
        if 'sncosmo' in config['object_types']:
            self.cat_cxt.register_source_type('sncosmo',
                                              object_class=SncosmoObject)
        if 'star' in config['object_types']:
            self.cat_cxt.register_source_type('star',
                                              object_class=StarObject)
        if 'galaxy' in config['object_types']:
            self.cat_cxt.register_source_type('galaxy',
                                              object_class=GalaxyObject)

    @property
    def observed_sed_factory(self):
        return self._observed_sed_factory

    @property
    def extinguisher(self):
        return self._extinguisher

    @property
    def raw_config(self):
        '''
        Return config, typically uploaded from yaml.
        '''
        return self._config

    # One might check that the config is complete and well-formed
    #  - for example require certain keys, such as catalog_name,
    #  data_file_type, area_parition, root_directory, object_types -
    # to exist, and check that that the data directory (value of
    # root_directory) exists.
    def _validate_config(self):
        pass

    def _find_all_hps(self):
        '''
        For each healpix with files matching pattern in the directory,
        update _hp_info as needed to keep track of all files for that healpix
        and the object types included in those files.

        Returns
        -------
        Set of healpix pixels with at least one file in the directory

        '''
        # If major organization is by healpix, healpix # could be in
        # subdirectory name.  Otherwise there may be no subdirectories
        # or data may be organized by component type.
        # Here only handle case where data files are directly in root dir
        files = os.listdir(self._cat_dir)
        o_types = self._config['object_types']

        hp_set = set()
        for f in files:
            for ot in o_types:
                # find all keys containing the string 'file_template'
                template_keys = [k for k in o_types[ot] if 'file_template' in k]
                for k in template_keys:
                    m = re.fullmatch(o_types[ot][k], f)
                    if m:
                        hp = int(m['healpix'])
                        hp_set.add(hp)

                        if hp not in self._hp_info:
                            self._hp_info[hp] = {'files' : {f : None},
                                                 'object_types' : {ot : [f]}}
                        else:
                            this_hp = self._hp_info[hp]
                            # Value of 'object_types' is now a list
                            if f not in this_hp['files'] :
                                this_hp['files'][f] = None
                            if ot in this_hp['object_types']:
                                this_hp['object_types'][ot].append(f)
                            else:
                                this_hp['object_types'][ot] = [f]
        return hp_set

    def get_hps_by_region(self, region, object_type='galaxy'):
        '''
        Parameters
        ----------
        Region can be a box (named 4-tuple (min-ra, max-ra, min-dec, max-dec)),
        a circle (named 3-tuple (ra, dec, radius)) or of type
        PolygonalRegion.
        Catalog area partition must be by healpix

        Returns
        -------
        Set of healpixels intersecting the region
        '''
        # If area_partition doesn't use healpix, raise exception
        if self._global_partition is None:
            partition = self._config['object_types'][object_type]['area_partition']
        else:
            partition = self._global_partition
        return _get_intersecting_hps(partition['ordering'], partition['nside'],
                                     region)

    def get_object_type_names(self):
        '''
        Returns
        -------
        All object type names in the catalog's config
        '''
        return set(self._config['object_types'].keys())

    # Add more functions to return parts of config of possible interest
    # to user

    def toplevel_only(self, object_types):
        '''
        Parameters
        ----------
        object_types     Set of object type names
        Remove object types with a parent.  Add in the parent.

        Return the resulting set
        '''
        objs_copy = set(object_types)
        for obj in object_types:
            parent = self._config.get_object_parent(obj)
            if parent is not None:
                objs_copy.remove(obj)
                objs_copy.add(parent)
        return objs_copy

    def get_objects_by_region(self, region, obj_type_set=None, mjd=None):
        '''
        Parameters
        ----------
        region         region is a named tuple(may be box or circle)
                       or object of type PolygonalRegion
        obj_type_set   Return only these objects. Defaults to all available
        mjd            MJD of observation epoch.

        Returns
        -------
        ObjectList containing sky objects visible in the region
        [at the specified time]
        '''
        # Take intersection of obj_type_list and available object types
        # Determine healpix intersecting the region

        if self.verbose:
            print("Region ", region)
            print("obj_type_set ", obj_type_set)
        # This must be done per object type
        # if self._config['area_partition']['type'] == 'healpix':
        #     hps = self.get_hps_by_region(region)

        # otherwise raise a not-supported exception

        object_list = ObjectList()
        if obj_type_set is None:
            obj_types = self.get_object_type_names()
        else:
            obj_types = self.get_object_type_names().intersection(obj_type_set)
        obj_types = self.toplevel_only(obj_types)

        for ot in obj_types:
            new_list = self.get_object_type_by_region(region, ot, mjd=mjd)
            object_list.append_object_list(new_list)

        return object_list

    def get_object_type_by_region(self, region, object_type, mjd=None):
        '''
        Parameters
        ----------
        region        box, circle or PolygonalRegion. Supported region
                      types made depend on object_type
        object_type   known object type without parent
        mjd           MJD of observation epoch.

        Returns all objects found
        '''

        out_list = ObjectList()
        if self._global_partition is not None:
            partition = self._global_partition
        else:
            partition = self._config['object_types'][object_type]['area_partition']

        coll_type = self.cat_cxt.lookup_collection_type(object_type)
        if coll_type is not None:
            out_list.append_collection(coll_type.load_collection(region, self,
                                                                 mjd=mjd))
            return out_list

        if partition != 'None':
            if partition['type'] == 'healpix':
                hps = self.get_hps_by_region(region, object_type)
                for hp in hps:
                    c = self.get_object_type_by_hp(hp, object_type, region, mjd)
                    if len(c) > 0:
                        out_list.append_object_list(c)
                return out_list
        else:
            raise NotImplementedError(f'Unsupported object type {object_type}')

    def get_object_type_by_hp(self, hp, object_type, region=None, mjd=None):
        object_list = ObjectList()

        #  Do we need to check more specifically by object type?
        if hp not in self._hp_info:
            print(f'WARNING: In SkyCatalog.get_object_type_by_hp healpixel {hp} intersects region but has no catalog file')
            return object_list

        if object_type == 'galaxy':
            COLUMNS = ['galaxy_id', 'ra', 'dec']
            id_name = 'galaxy_id'
        elif object_type in ['star', 'sncosmo']:
            COLUMNS = ['object_type', 'id', 'ra', 'dec']
            id_name = 'id'
        else:
            raise NotImplementedError(f'Unsupported object type {object_type}')

        if self.verbose:
            print('Working on healpix pixel ', hp)
        rdr_ot = dict()   # maps readers to set of object types it reads

        if 'file_template' in self._config['object_types'][object_type]:
            f_list = self._hp_info[hp]['object_types'][object_type]
        elif 'parent' in self._config['object_types'][object_type]:
            f_list = self._hp_info[hp]['object_types'][self._config['object_types'][ot]['parent']]

        for f in f_list:
            if self._hp_info[hp]['files'][f] is None:            # no reader yet
                full_path = os.path.join(self._cat_dir, f)
                the_reader = ParquetReader(full_path, mask=None)
                self._hp_info[hp]['files'][f] = the_reader
            else:
                the_reader = self._hp_info[hp]['files'][f]
                # associate object type with this reader
            if the_reader in rdr_ot:
                rdr_ot[the_reader].add(object_type)
            else:
                rdr_ot[the_reader] = set([object_type])

        # Find readers needed to get minimal columns for our object type
        the_readers = []
        for rdr in rdr_ot:
            if object_type in rdr_ot[rdr]:
                the_readers.append(rdr)

        # Unfortunately galaxies and point sources can't be handled quite
        # the same way since we need to read an extra column for pointsources
        # There will have to be some "if galaxy... else ... code"
        for rdr in the_readers:
            if 'ra' not in rdr.columns:
                continue

            # Make a collection for each row group
            for rg in range(rdr.n_row_groups):
                arrow_t = rdr.read_columns(COLUMNS, None, rg)
                if object_type == 'galaxy':
                    ra_c, dec_c, id_c, mask = _compress_via_mask(arrow_t,
                                                                 id_name,
                                                                 region)
                    if ra_c is not None:
                        new_collection = ObjectCollection(ra_c, dec_c, id_c,
                                                          'galaxy', hp, self,
                                                          region=region,
                                                          mjd=mjd,
                                                          mask=mask,
                                                          readers=the_readers,
                                                          row_group=rg)
                        object_list.append_collection(new_collection)

                else:
                    ra_c, dec_c, id_c, object_type_c, mask =\
                        _compress_via_mask(arrow_t, id_name, region,
                                           galaxy=False)
                    if ra_c is not None and object_type_c[0] == object_type:
                        new_collection = ObjectCollection(ra_c, dec_c, id_c,
                                                          object_type_c[0], hp,
                                                          self, region=region,
                                                          mjd=mjd,
                                                          mask=mask,
                                                          readers=the_readers,
                                                          row_group=rg)
                        object_list.append_collection(new_collection)

        return object_list


    # For generator version, do this a row group at a time
    #    but if region cut leaves too small a list, read more rowgroups
    #    to achieve a reasonable size list (or exhaust the file)
    def get_object_iterator_by_hp(self, hp, obj_type_set=None,
                                  max_chunk=None, mjd=None):
        '''
        Parameters
        ----------
        mjd            MJD of observation epoch
        hp             A healpix id
        obj_type_set   Return only these objects. Defaults to all available
        max_chunk      If specified, iterator will return no more than this
                       number of objections per iteration
        Returns
        -------
        An iterator
        '''
        pass


def open_catalog(config_file, mp=False, skycatalog_root=None, verbose=False):
    '''
    Parameters
    ----------
    config_file   yaml file containing config
    skycatalog_root optional override of skycatalog_root. This value may also
                    be supplied by setting the environment variable
                    SKYCATALOG_ROOT.  Precedence is 1) argument
                    2) environment variable 3) value in config file for
                    key skycatalog_root.  However set, this value
                    joined to value of the key catalog_dir will be used
                    to find the catalog data.

    Returns
    -------
    SkyCatalog
    '''
    # Get LSST bandpasses in case we need to compute fluxes
    band_passes = load_lsst_bandpasses()
    with open(config_file) as f:
        return SkyCatalog(yaml.safe_load(f), skycatalog_root=skycatalog_root,
                          mp=mp, verbose=verbose)

if __name__ == '__main__':
    import time
    cfg_file_name = 'skyCatalog.yaml'
    skycatalog_root = os.getenv('SKYCATALOG_ROOT')
    catalog_dir = 'reorg'
    if len(sys.argv) > 1:
        catalog_dir = sys.argv[1]
    if len(sys.argv) > 2:
        cfg_file_name = sys.argv[2]

    cfg_file = os.path.join(skycatalog_root, catalog_dir, cfg_file_name)

    write_sed = False
    if len(sys.argv) > 3:
        write_sed = True

    # For tract 3828
    #   55.73604 < ra < 57.563452
    #  -37.19001 < dec < -35.702481


    cat = open_catalog(cfg_file, skycatalog_root=skycatalog_root)
    hps = cat._find_all_hps()
    print('Found {} healpix pixels '.format(len(hps)))
    for h in hps: print(h)

    ra_min_tract = 55.736
    ra_max_tract = 57.564
    dec_min_tract = -37.190
    dec_max_tract = -35.702
    ##ra_min_small = 56.0
    ##ra_max_small = 56.2
    ra_min_small = 55.9
    ra_max_small = 56.1
    dec_min_small = -36.2
    dec_max_small = -36.0

    sed_fmt = 'lambda: {:.1f}  f_lambda: {:g}'

    rgn = Box(ra_min_small, ra_max_small, dec_min_small, dec_max_small)
    vertices = [(ra_min_small, dec_min_small), (ra_min_small, dec_max_small),
                (ra_max_small, dec_max_small), (ra_max_small, dec_min_small)]
    rgn_poly = PolygonalRegion(vertices_radec=vertices)

    intersect_hps = _get_intersecting_hps('ring', 32, rgn)

    print("For region ", rgn)
    print("intersecting pixels are ", intersect_hps)

    intersect_poly_hps = _get_intersecting_hps('ring', 32, rgn_poly)
    print("For region ", rgn_poly)
    print("intersecting pixels are ", intersect_poly_hps)

    at_slac = os.getenv('HOME').startswith('/sdf/home/')
    if not at_slac:
        obj_types = {'star', 'galaxy', 'sncosmo'}
    else:
        obj_types = {'star', 'galaxy', 'sncosmo', 'gaia_star'}

    print('Invoke get_objects_by_region with box region, no gaia')
    t0 = time.time()
    object_list = cat.get_objects_by_region(rgn,
                                            obj_type_set={'star','galaxy',
                                                          'sncosmo'})
    t_done = time.time()
    print('Took ', t_done - t0)
                                            ##### temporary obj_type_set={'galaxy', 'star'} )
    #                                        obj_type_set=set(['galaxy']) )
    # Try out get_objects_by_hp with no region
    #colls = cat.get_objects_by_hp(9812, None, set(['galaxy']) )

    print('Number of collections returned for box:  ', object_list.collection_count)
    print('Object count for box: ', len(object_list))

    print('Invoke get_objects_by_region with polygonal region')
    t0 = time.time()
    object_list_poly = cat.get_objects_by_region(rgn_poly,
                                                 obj_type_set=obj_types)
    t_done = time.time()
    print('Took ', t_done - t0)

    print('Number of collections returned for polygon:  ',
          object_list_poly.collection_count)
    assert(object_list.collection_count == object_list_poly.collection_count)
    print('Object count for polygon: ', len(object_list_poly))

    fudge = 5
    assert(len(object_list_poly) > len(object_list) - fudge)
    assert(len(object_list_poly) < len(object_list) + fudge)

    print('Now try inscribed polygon which is not a box')
    ra_avg = (ra_min_small + ra_max_small) * 0.5
    dec_avg = (dec_min_small + dec_max_small) * 0.5
    diamond_vertices = [(ra_min_small, dec_avg), (ra_avg, dec_max_small),
                        (ra_max_small, dec_avg), (ra_avg, dec_min_small)]
    rgn_diamond = PolygonalRegion(vertices_radec=diamond_vertices)

    intersect_diamond_hps = _get_intersecting_hps('ring', 32, rgn_diamond)
    print("for diamond region ", rgn_diamond)
    print("intersecting pixels are ", intersect_diamond_hps)

    print('Invoke get_objects_by_region with diamond region')
    t0 = time.time()
    object_list_diamond = cat.get_objects_by_region(rgn_diamond,
                                                    obj_type_set=obj_types)
    t_done = time.time()
    print('Took ', t_done - t0)

    print('Number of collections returned for diamond:  ',
          object_list_diamond.collection_count)
    print('Object count for diamond: ', len(object_list_diamond))


    #### TEMP FOR DEBUGGING
    ### exit(0)

    # For now SIMS_SED_LIBRARY_DIR is undefined at SLAC, making it impossible
    # to get SEDs for stars. So (crudely) determine whether or not
    # we're running at SLAC

    colls = object_list.get_collections()
    got_a_sed = False
    for c in colls:
        n_obj = len(c)
        print("For hpid ", c.get_partition_id(), "found ", n_obj, " objects")
        if (n_obj) < 1:
            continue
        print("First object: ")
        print(c[0], '\nid=', c[0].id, ' ra=', c[0].ra, ' dec=', c[0].dec,
              ' belongs_index=', c[0]._belongs_index,
              ' object_type: ', c[0].object_type )

        if (n_obj < 3):
            continue
        print("Slice [1:3]")
        slice13 = c[1:3]
        for o in slice13:
            print('id=',o.id, ' ra=',o.ra, ' dec=',o.dec, ' belongs_index=',
                  o._belongs_index,  ' object_type: ', o.object_type)
            print(o.object_type)
            if o.object_type == 'star':
                if not at_slac:
                    print(o.get_instcat_entry())
                    sed, magnorm = o._get_sed()
                    print('For star magnorm: ', magnorm)
                    if magnorm < 1000:
                        print('Length of sed: ', len(sed.wave_list))
            elif o.object_type == 'sncosmo':
                print(o.get_instcat_entry())
            elif o.object_type == 'galaxy':
                for cmp in ['disk', 'bulge', 'knots']:
                    print(cmp)
                    if cmp in o.subcomponents:
                        # broken for galaxies currently
                        ###print(o.get_instcat_entry(component=cmp))
                        sed, _ = o._get_sed(cmp)
                        if sed:
                            print('Length of sed table: ', len(sed.wave_list))
                            if not got_a_sed:
                                got_a_sed = True
                                th = o.get_native_attribute(f'sed_val_{cmp}')
                                print('Tophat values: ', th)
                                sed, _ = o._get_sed(component=cmp)
                                print('Simple sed wavelengths:')
                                print(sed.wave_list)
                                print('Simple sed values:')
                                print([sed(w) for w in sed.wave_list])
                                if write_sed:
                                    o.write_sed('simple_sed.txt', component=cmp)
                                sed_fine, _ = o._get_sed(component=cmp,
                                                        resolution=1.0)
                                print('Bin width = 1 nm')
                                print('Initial wl values', sed_fine.wave_list[:20])
                                print('Start at bin 100', sed_fine.wave_list[100:120])
                                print('Initial values')
                                print([sed_fine(w) for w in sed_fine.wave_list[:20]])
                                print('Start at bin 100')
                                print([sed_fine(w) for w in sed_fine.wave_list[100:120]])
                        else:
                            print('All-zero sed')

                # Try out old wrapper functions
                print("\nget_dust:")
                i_av, i_rv, g_av, g_rv = o._get_dust()
                print(f'i_av={i_av} i_rv={i_rv} g_av={g_av} g_rv={g_rv}')
                print("\nget_wl_params")
                g1, g2, mu = o.get_wl_params()
                print(f'g1={g1} g2={g2} mu={mu}')
                print("\nget_gsobject_components. Keys of returned dict:")
                gs_dict = o.get_gsobject_components()
                print(gs_dict.keys())
                print("\nget_observer_sed_components.  Keys of returned dict:")
                o_seds = o.get_observer_sed_components()
                print(o_seds.keys())

                f = o.get_LSST_flux('i')
                print(f'Flux for i bandpass: {f}')
                fluxes = o.get_LSST_fluxes()
                for k,v in fluxes.items():
                    print(f'Bandpass {k} has flux {v}')

        if n_obj > 200:
            print("Object 200")
            print(c[200], '\nid=', c[200].id, ' ra=', c[200].ra, ' dec=',
                  c[200].dec,
                  ' belongs_index=', c[200]._belongs_index)
        if n_obj > 163997:
            slice_late = c[163994:163997]
            print('\nobjects indexed 163994 through 163996')
            for o in slice_late:
                print('id=',o.id, ' ra=',o.ra, ' dec=',o.dec, ' belongs_index=',
                      o._belongs_index)

    print('Total object count: ', len(object_list))

    if len(object_list) == 0:
        print('Empty object list. All done')
        exit(0)

    obj = object_list[0]
    print("Type of element in object_list:", type(obj))

    if object_list[0].object_type == 'galaxy':
        redshift0 = object_list[0].get_native_attribute('redshift')
        print('First redshift: ', redshift0)


    sum = 0
    for obj in object_list:
        sum = sum + 1
        if sum > 7216:
            print("Sum is now ", sum)
            print("obj id: ", obj.id)
        if sum > 7220:
            break

    print(f'Object list len:  {len(object_list)}')
    print(f'Objects found with "in":  {sum}')

    if len(object_list) < 5:
        print('Very short object list (< 5 elements).  done')
        exit(0)

    segment = object_list[2:5]
    print('Information for slice 2:5 ')
    for o in segment:
        print(f'object {o.id} of type {o.object_type} belongs to collection {o._belongs_to}')

    if len(object_list) < 304:
        print('Object list len < 304.  All done')
        exit(0)
    print('\nInformation for slice 285:300')
    segment = object_list[285:300]
    for o in segment:
        print(f'object {o.id} of type {o.object_type} belongs to collection {o._belongs_to}')

    #ixes = ([3,5,8],)
    ixes = (np.array([3,5,8, 300, 303]),)
    print(f'\nObjects with indexes {ixes[0]}')
    for o in object_list[ixes]:
        print(o.id)
    print(f'\nObjects in slice [3:9]')
    for o in object_list[3:9]:
        print(o.id)
    print(f'\nObjects in slice [300:304]')
    for o in object_list[300:304]:
        print(o.id)

    print('all done')



import os
import re
from astropy import units as u
from astropy.cosmology import FlatLambdaCDM
import astropy.constants
import h5py
import pandas as pd

import numpy as np
import numpy.ma as ma
from dust_extinction.parameter_averages import F19
import galsim

__all__ = ['TophatSedFactory', 'MilkyWayExtinction', 'get_star_sed_path']
class TophatSedFactory:
    '''
    Used for modeling cosmoDC2 galaxy SEDs, which are represented with
    a small number of wide bins
    '''
    _clight = astropy.constants.c.to('m/s').value
    # Conversion factor below of cosmoDC2 tophat Lnu values to W/Hz comes from
    # https://github.com/LSSTDESC/gcr-catalogs/blob/master/GCRCatalogs/SCHEMA.md
    _to_W_per_Hz = 4.4659e13

    def __init__(self, th_definition, cosmology, delta_wl=0.001):
        # Get wavelength and frequency bin boundaries.
        if th_definition:
            bins = th_definition
            wl0 = [_[0] for _ in bins]

            # Find index of original bin which includes 500 nm == 5000 ang
            ix = -1
            for w in wl0:
                if w > 5000:
                    break
                ix += 1

            self._ix_500nm = ix

            wl0.append(bins[-1][0] + bins[-1][1])

            wl0 = 0.1*np.array(wl0)
            self.wl = np.array(wl0)
            self.nu = self._clight/(self.wl*1e-9)  # frequency in Hz

            # Also save version of wl where vertical rise is replaced by
            # steep slope
            wl_deltas = []
            for i in range(len(bins)):
                wl_deltas.extend((self.wl[i], self.wl[i+1] - delta_wl))

            # Prepend more bins which will have 0 value
            n_bins = int(wl0[0]) - 1
            pre_wl = [float(i) for i in range(n_bins)]

            wl_deltas = np.insert(wl_deltas, 0, pre_wl)

            # Also make a matching array of 0 values
            self.pre_val = [0.0 for i in range(n_bins)]

            self._wl_deltas = wl_deltas
            self._wl_deltas_u_nm = wl_deltas*u.nm

        # Create a FlatLambdaCDM cosmology from a dictionary of input
        # parameters.  This code is based on/borrowed from
        # https://github.com/LSSTDESC/gcr-catalogs/blob/master/GCRCatalogs/cosmodc2.py#L128
        cosmo_astropy_allowed = FlatLambdaCDM.__init__.__code__.co_varnames[1:]
        cosmo_astropy = {k: v for k, v in cosmology.items()
                         if k in cosmo_astropy_allowed}
        self.cosmology = FlatLambdaCDM(**cosmo_astropy)

    # Useful for getting magnorm from f_nu values
    @property
    def ix_500nm(self):
        return self._ix_500nm

    @property
    def wl_deltas(self):
        return self._wl_deltas

    @property
    def wl_deltas_u_nm(self):
        return self._wl_deltas_u_nm

    def dl(self, z):
        """
        Return the luminosity distance in units of meters.
        """
        # Conversion factor from Mpc to meters (obtained from pyccl).
        MPC_TO_METER = 3.085677581491367e+22
        return self.cosmology.luminosity_distance(z).value*MPC_TO_METER

    def create(self, Lnu, redshift_hubble, redshift, resolution=None):
        '''
        Given tophat values from cosmoDC2 produce redshifted sed.
        Does not apply extinction.
        '''
        # Compute Llambda in units of W/nm
        Llambda = (Lnu*self._to_W_per_Hz*(self.nu[:-1] - self.nu[1:])
                   /(self.wl[1:] - self.wl[:-1]))

        # Fill the arrays for the galsim.LookupTable.   Prepend
        # zero-valued bins down to mix extinction wl to handle redshifts z > 2.
        my_Llambda = []
        my_Llambda += self.pre_val
        for i in range(len(Llambda)):
            # Dealt with wl already in __init__
            my_Llambda.extend((Llambda[i], Llambda[i]))

        # Convert to (unredshifted) flux given redshift_hubble.
        flambda = np.array(my_Llambda)/(4.0*np.pi*self.dl(redshift_hubble)**2)

        # Convert to cgs units
        flambda *= (1e7/1e4)  # (erg/joule)*(m**2/cm**2)

        # Create the lookup table.
        lut = galsim.LookupTable(self.wl_deltas, flambda, interpolant='nearest')

        if resolution:
            wl_min = min(self.wl_deltas)
            wl_max = max(self.wl_deltas)
            wl_res = np.linspace(wl_min, wl_max, int((wl_max - wl_min)/resolution))
            flambda_res = [lut(wl) for wl in wl_res]
            lut = galsim.LookupTable(wl_res, flambda_res, interpolant='linear')

        # Create the SED object and apply redshift.
        sed = galsim.SED(lut, wave_type='nm', flux_type='flambda')\
                    .atRedshift(redshift)

        return sed

    def magnorm(self, tophat_values, z_H):
        one_Jy = 1e-26  # W/Hz/m**2
        Lnu = tophat_values[self.ix_500nm]*self._to_W_per_Hz  # convert to W/Hz
        Fnu = Lnu/4/np.pi/self.dl(z_H)**2
        with np.errstate(divide='ignore', invalid='ignore'):
            return -2.5*np.log10(Fnu/one_Jy) + 8.90

class MilkyWayExtinction:
    '''
    Applies extinction to a SED
    '''
    def __init__(self, delta_wl=0.1, mwRv=3.1, eps=1e-7):
        """
        Parameters
        ----------
        sed_factory: (remove)
        ext_bin_width: (rename to delta_wl?)
        delta_wl : float [0.1]
            Wavelength sampling of the extinction function in nm
        mwRv : float [3.1]
            Parameter describing the shape of the Milky Way extinction
            curve.
        eps : float [1e-7]
            Small numerical offset to avoid out-of-range errors in
            the wavelength array passed to the dust_extinction code.
        """
        # Wavelength sampling for the extinction function. F19.x_range
        # is in units of 1/micron so convert to nm.  The eps value
        # is needed to avoid numerical noise at the end points causing
        # out of range errors detected by the dust_extinction code.
        wl_min = 1e3/F19.x_range[1] + eps
        wl_max = 1e3/F19.x_range[0] - eps
        npts = int((wl_max - wl_min)/delta_wl)
        self.wls = np.linspace(wl_min, wl_max, npts)
        self.extinction = F19(Rv=mwRv)

    def extinguish(self, sed, mwAv):
        ext = self.extinction.extinguish(self.wls*u.nm, Av=mwAv)
        lut = galsim.LookupTable(self.wls, ext, interpolant='linear')
        mw_ext = galsim.SED(lut, wave_type='nm', flux_type='1')
        sed = sed*mw_ext
        return sed


_standard_dict = {'lte' : 'starSED/phoSimMLT',
                  'bergeron' : 'starSED/wDs',
                  'km|kp' : 'starSED/kurucz'}

def get_star_sed_path(filename, name_to_folder=_standard_dict):
    '''
    Return numpy array of full paths relative to SIMS_SED_LIBRARY_DIR,
    given filenames

    Parameters
    ----------
    filename       list of strings. Usually full filename but may be missing final ".gz"
    name_to_folder dict mapping regular expression (to be matched with
                   filename) to relative path for containing directory

    Returns
    -------
    Full path for file, relative to SIMS_SED_LIBRARY_DIR
    '''

    compiled = { re.compile(k) : v for (k, v) in name_to_folder.items()}

    path_list = []
    for f in filename:
        m = None
        matched = False
        for k,v in compiled.items():
            f = f.strip()
            m = k.match(f)
            if m:
                p = os.path.join(v, f)
                if not p.endswith('.gz'):
                    p = p + '.gz'
                path_list.append(p)
                matched = True
                break

        if not matched:
            raise ValueError(f'get_star_sed_path: Filename {f} does not match any known patterns')
    return np.array(path_list)



import pyarrow as pa
import logging

__all__ = ['make_galaxy_schema', 'make_galaxy_flux_schema',
           'make_pointsource_schema', 'make_star_flux_schema']

# This schema is not the same as the one taken from the data,
# probably because of the indexing in the schema derived from a pandas df.
def make_galaxy_schema(logname, sed_subdir=False, knots=True):
    fields = [pa.field('galaxy_id', pa.int64()),
              pa.field('ra', pa.float64() , True),
##                       metadata={"units" : "radians"}),
              pa.field('dec', pa.float64() , True),
##                       metadata={"units" : "radians"}),
              pa.field('redshift', pa.float64(), True),
              pa.field('redshift_hubble', pa.float64(), True),
              pa.field('peculiar_velocity', pa.float64(), True),
              pa.field('shear_1', pa.float64(), True),
              pa.field('shear_2', pa.float64(), True),
              pa.field('convergence', pa.float64(), True),
              pa.field('size_bulge_true', pa.float32(), True),
              pa.field('size_minor_bulge_true', pa.float32(), True),
              pa.field('sersic_bulge', pa.float32(), True),
              pa.field('size_disk_true', pa.float32(), True),
              pa.field('size_minor_disk_true', pa.float32(), True),
              pa.field('sersic_disk', pa.float32(), True),
##              pa.field('position_angle_unlensed', pa.float64(), True),
              # Depending on value of --dc2-like option, value for
              # ellipticity_2_true column will differ
              pa.field('ellipticity_1_disk_true', pa.float64(), True),
              pa.field('ellipticity_2_disk_true', pa.float64(), True),
              pa.field('ellipticity_1_bulge_true', pa.float64(), True),
              pa.field('ellipticity_2_bulge_true', pa.float64(), True),
              pa.field('sed_val_bulge',
                       pa.list_(pa.float64()), True),
              pa.field('sed_val_disk',
                       pa.list_(pa.float64()), True),
              pa.field('bulge_magnorm', pa.float64(), True),
              pa.field('disk_magnorm', pa.float64(), True),
              pa.field('MW_rv', pa.float32(), True),
              pa.field('MW_av', pa.float32(), True)]
    logger = logging.getLogger(logname)
    if knots:
        logger.debug("knots requested")
        fields.append(pa.field('sed_val_knots',
                               pa.list_(pa.float64()), True))
        ### For sizes API can alias to disk sizes
        ###  position angle, shears and convergence are all
        ###  galaxy-wide quantities.
        fields.append(pa.field('n_knots', pa.float32(), True))
        fields.append(pa.field('knots_magnorm', pa.float64(), True))

    if sed_subdir:
        fields.append(pa.field('bulge_sed_file_path', pa.string(), True))
        fields.append(pa.field('disk_sed_file_path', pa.string(), True))

    debug_out = ''
    for f in fields:
        debug_out += f'{f.name}\n'
    logger.debug(debug_out)
    return pa.schema(fields)

def make_galaxy_flux_schema(logname):
    '''
    Will make a separate parquet file with lsst flux for each band
    and galaxy id for joining with the main galaxy file
    '''
    logger = logging.getLogger(logname)
    logger.debug('Creating galaxy flux schema')
    fields = [pa.field('galaxy_id', pa.int64()),
              # should flux fields be named e.g. lsst_cmodel_flux_u?
              pa.field('lsst_flux_u', pa.float32() , True),
              pa.field('lsst_flux_g', pa.float32() , True),
              pa.field('lsst_flux_r', pa.float32() , True),
              pa.field('lsst_flux_i', pa.float32() , True),
              pa.field('lsst_flux_z', pa.float32() , True),
              pa.field('lsst_flux_y', pa.float32() , True)]
    return pa.schema(fields)

def make_star_flux_schema(logname):
    '''
    Will make a separate parquet file with lsst flux for each band
    and id for joining with the main star file
    '''
    logger = logging.getLogger(logname)
    logger.debug('Creating star flux schema')
    fields = [pa.field('id', pa.string()),
              pa.field('lsst_flux_u', pa.float32() , True),
              pa.field('lsst_flux_g', pa.float32() , True),
              pa.field('lsst_flux_r', pa.float32() , True),
              pa.field('lsst_flux_i', pa.float32() , True),
              pa.field('lsst_flux_z', pa.float32() , True),
              pa.field('lsst_flux_y', pa.float32() , True)]
    return pa.schema(fields)

def make_pointsource_schema():
    '''
    Ultimately should handle stars both static and variable, SN, and AGN
    For now add everything needed for SN and put in some additional
    star fields, but not structs for star variability models
    '''

    salt2_fields = [
        pa.field('z', pa.float64(), True),
        pa.field('t0', pa.float64(), True),
        pa.field('x0', pa.float64(), True),
        pa.field('x1', pa.float64(), True),
        pa.field('c', pa.float64(), True)]
    fields = [pa.field('object_type', pa.string(), False),
              pa.field('id', pa.string(), False),
              pa.field('ra', pa.float64(), False),
              pa.field('dec', pa.float64(), False),
              pa.field('host_galaxy_id', pa.int64(), True),
              pa.field('magnorm', pa.float64(), True),
              pa.field('sed_filepath', pa.string(), True),
              pa.field('MW_rv', pa.float32(), True),
              pa.field('MW_av', pa.float32(), True),
              pa.field('mura', pa.float64(), True),
              pa.field('mudec', pa.float64(), True),
              pa.field('radial_velocity', pa.float64(), True),
              pa.field('parallax', pa.float64(), True),
              pa.field('variability_model', pa.string(), True),
              pa.field('salt2_params', pa.struct(salt2_fields), True)
    ]
    return pa.schema(fields)

def make_pointsource_flux_schema(logname):
    '''
    Will make a separate parquet file with lsst flux for each band
    and id for joining with the main star file.
    For static sources mjd field could be -1. Or the field could be
    made nullable.
    '''
    logger = logging.getLogger(logname)
    logger.debug('Creating pointsource flux schema')
    fields = [pa.field('id', pa.string()),
              pa.field('lsst_flux_u', pa.float32() , True),
              pa.field('lsst_flux_g', pa.float32() , True),
              pa.field('lsst_flux_r', pa.float32() , True),
              pa.field('lsst_flux_i', pa.float32() , True),
              pa.field('lsst_flux_z', pa.float32() , True),
              pa.field('lsst_flux_y', pa.float32() , True),
              pa.field('mjd', pa.float64() , True)]
    return pa.schema(fields)

    return pa.schema(fields)
